@@ -4,28 +4,32 @@ import datetime
 import decimal
 import uuid
 from collections import OrderedDict
-from typing import List, Type, Dict, Any, Tuple, Mapping
+from typing import Any, Dict, Generic, Mapping, List, Tuple, Type, TypeVar
 
+import rest_framework.fields
+import rest_framework.serializers
 from django.core.exceptions import ImproperlyConfigured
 from django.db.models import Model
 from django.utils.functional import cached_property
-import rest_framework.fields
-import rest_framework.serializers
-from rest_framework.relations import PrimaryKeyRelatedField, HyperlinkedRelatedField
+from rest_framework.fields import empty
+from rest_framework.relations import HyperlinkedRelatedField, PrimaryKeyRelatedField
 from rest_framework.utils.field_mapping import get_relation_kwargs
 
-from rest_framework_dataclasses import field_utils, typing_utils
-from rest_framework_dataclasses.field_utils import DataclassDefinition, get_dataclass_definition, TypeInfo
+from rest_framework_dataclasses import fields, field_utils, typing_utils
+from rest_framework_dataclasses.field_utils import get_dataclass_definition, DataclassDefinition, TypeInfo
+from rest_framework_dataclasses.types import Dataclass
 
 
 # Define some types to make type hinting more readable
-SerializerField = rest_framework.fields.Field
+AnyT = TypeVar('AnyT')
 KWArgs = Dict[str, Any]
+SerializerField = rest_framework.fields.Field
 SerializerFieldDefinition = Tuple[Type[SerializerField], KWArgs]
+T = TypeVar('T', bound=Dataclass)
 
 
 # noinspection PyMethodMayBeStatic
-class DataclassSerializer(rest_framework.serializers.Serializer):
+class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
     """
     A `DataclassSerializer` is just a regular `Serializer`, except that:
 
@@ -48,12 +52,16 @@ class DataclassSerializer(rest_framework.serializers.Serializer):
         float:              rest_framework.fields.FloatField,
         bool:               rest_framework.fields.BooleanField,
         str:                rest_framework.fields.CharField,
-        decimal.Decimal:    rest_framework.fields.DecimalField,
+        decimal.Decimal:    fields.DefaultDecimalField,
         datetime.date:      rest_framework.fields.DateField,
         datetime.datetime:  rest_framework.fields.DateTimeField,
         datetime.time:      rest_framework.fields.TimeField,
         datetime.timedelta: rest_framework.fields.DurationField,
         uuid.UUID:          rest_framework.fields.UUIDField,
+        # Note that these are only used for class typehints, typing.Dict and typing.List are handled by
+        # build_composite_field().
+        dict:               rest_framework.fields.DictField,
+        list:               rest_framework.fields.ListField
     }
     serializer_related_field = PrimaryKeyRelatedField
 
@@ -99,48 +107,61 @@ class DataclassSerializer(rest_framework.serializers.Serializer):
 
     # Default `create` and `update` behavior...
 
-    def instantiate_data_dictionaries(self, validated_data, instance=None):
+    def _strip_empty_sentinels(self, x: AnyT) -> AnyT:
         """
-        Instantiate the values in a deserialized data dictionaries to their respective classes (for now only nested
-        dataclasses).
+        Recursively strip the `empty` sentinel values from dataclasses, replacing them with their fields default value.
         """
-        # I'm not sure this is actually the best way to deserialize into nested dataclasses. The cleanest way seems to
-        # be overriding to_internal_value(), but the top-level serializer must return a dictionary from that method. We
-        # could split nested dataclasses off into a separate DataclassField (which could in general clean-up the code a
-        # bit), but that breaks specifying nested serializers using a single class. Let's use this ugly hack until I can
-        # think of something better.
+        if dataclasses.is_dataclass(x) and not isinstance(x, type):
+            values = {field.name: self._strip_empty_sentinels(getattr(x, field.name)) for field in dataclasses.fields(x)
+                      if getattr(x, field.name) is not empty}
+            return type(x)(**values)
+        if isinstance(x, list):
+            return [self._strip_empty_sentinels(item) for item in x]
+        if isinstance(x, dict):
+            return {key: self._strip_empty_sentinels(value) for key, value in x.items()}
+        return x
 
-        ret = {}
-        for attr, value in validated_data.items():
-            field = self.fields[attr]
+    def create(self, validated_data: T) -> T:
+        # Create a new instance with all the `empty` sentinel values replaced by their defaults.
+        return self._strip_empty_sentinels(validated_data)
 
-            if value is None:
-                # If the value is None, we don't care about mapping it to the correct dataclass type.
-                pass
-            elif isinstance(field, DataclassSerializer):
-                value = (field.update(getattr(instance, attr), value)
-                         if instance and getattr(instance, attr) else field.create(value))
-            elif (isinstance(field, rest_framework.fields.ListField) and
-                  isinstance(field.child, DataclassSerializer) or
-                  isinstance(field, rest_framework.serializers.ListSerializer) and
-                  isinstance(field.child, DataclassSerializer)):
-                value = [field.child.create(item) if item else None for item in value]
-            elif isinstance(field, rest_framework.fields.DictField) and isinstance(field.child, DataclassSerializer):
-                value = {key: field.child.create(item) if item else None for key, item in value.items()}
+    def update(self, instance: T, validated_data: T) -> T:
+        for name in self.dataclass_definition.fields.keys():
+            # Only update fields that have been set in the serialized representation.
+            value = getattr(validated_data, name)
+            if value is not empty:
+                setattr(instance, name, self._strip_empty_sentinels(value))
 
-            ret[attr] = value
-
-        return ret
-
-    def create(self, validated_data):
-        dataclass_type = self.dataclass_definition.dataclass_type
-        return dataclass_type(**self.instantiate_data_dictionaries(validated_data))
-
-    def update(self, instance, validated_data):
-        for attr, value in self.instantiate_data_dictionaries(validated_data, instance).items():
-            setattr(instance, attr, value)
         return instance
 
+    def save(self, **kwargs) -> T:
+        assert hasattr(self, '_errors'), (
+            "You must call `.is_valid()` before calling `.save()`."
+        )
+
+        assert not self.errors, (
+            "You cannot call `.save()` on a serializer with invalid data."
+        )
+
+        assert not hasattr(self, '_data'), (
+            "You cannot call `.save()` after accessing `serializer.data`."
+            "If you need to access data before committing to the dataclass then "
+            "inspect 'serializer.validated_data' instead. "
+        )
+
+        validated_data = dataclasses.replace(self.validated_data, **kwargs)
+
+        if self.instance is not None:
+            self.instance = self.update(self.instance, validated_data)
+        else:
+            self.instance = self.create(validated_data)
+
+        assert self.instance is not None, (
+            '`update()` or `create()` did not return an object instance.'
+        )
+
+        return self.instance
+
     # Determine the fields to apply...
 
     def get_fields(self) -> Dict[str, SerializerField]:
@@ -214,18 +235,23 @@ class DataclassSerializer(rest_framework.serializers.Serializer):
         if fields is None or fields == rest_framework.serializers.ALL_FIELDS:
             fields = [rest_framework.serializers.ALL_FIELDS]
 
-        if rest_framework.serializers.ALL_FIELDS not in fields:
-            # If fields are explicitly specified, and it is not the all fields magic option, ensure that all included
-            # fields are valid.
-            for field_name in fields:
-                assert field_name in self.dataclass_definition.fields or field_name in declared_fields, (
-                    "The field '{field_name}' was included on serializer {serializer_class} in the `fields` option, "
-                    "but does not match any dataclass field."
-                    .format(field_name=field_name, serializer_class=self.__class__.__name__)
-                )
+        # For explicitly specified fields, ensure that they are valid.
+        for field_name in fields:
+            assert (
+                field_name == rest_framework.serializers.ALL_FIELDS or                         # all fields magic option
+                field_name in self.dataclass_definition.fields or                              # dataclass fields
+                field_name in declared_fields or                                               # declared fields
+                callable(getattr(self.dataclass_definition.dataclass_type, field_name, None))  # methods
+            ), (
+                "The field '{field_name}' was included on serializer {serializer_class} in the `fields` option, "
+                "but does not match any dataclass field."
+                .format(field_name=field_name, serializer_class=self.__class__.__name__)
+            )
 
-            # Also ensure that all declared fields are included. Do not require any fields that are declared in a parent
-            # class, in order to allow serializer subclasses to only include a subset of fields.
+        if rest_framework.serializers.ALL_FIELDS not in fields:
+            # If there are only explicitly specified fields, ensure that all declared fields are included. Do not
+            # require any fields that are declared in a parent class, in order to allow serializer subclasses to only
+            # include a subset of fields.
             required_field_names = set(declared_fields)
             for cls in self.__class__.__bases__:
                 required_field_names -= set(getattr(cls, '_declared_fields', []))
@@ -236,7 +262,7 @@ class DataclassSerializer(rest_framework.serializers.Serializer):
                     "included in the `fields` option."
                     .format(field_name=field_name, serializer_class=self.__class__.__name__)
                 )
-            return fields
+            return list(fields)
 
         # The field list now includes the magic all fields option, so replace it with the default field names.
         fields = list(fields)
@@ -263,8 +289,7 @@ class DataclassSerializer(rest_framework.serializers.Serializer):
 
         return fields
 
-    def get_default_field_names(self, declared_fields: Mapping[str, SerializerField]) \
-            -> List[str]:
+    def get_default_field_names(self, declared_fields: Mapping[str, SerializerField]) -> List[str]:
         """
         Return the default list of field names that will be used if the `Meta.fields` option is not specified.
         """
@@ -284,7 +309,9 @@ class DataclassSerializer(rest_framework.serializers.Serializer):
 
         # Determine the serializer field class and keyword arguments.
         if source in self.dataclass_definition.fields:
-            type_info = field_utils.get_type_info(self.dataclass_definition.field_types[source])
+            default_value = self.dataclass_definition.fields[source].default
+            default_type = type(default_value) if default_value is not None else None
+            type_info = field_utils.get_type_info(self.dataclass_definition.field_types[source], default_type)
             field_class, field_kwargs = self.build_typed_field(source, type_info, extra_kwargs)
         elif hasattr(self.dataclass_definition.dataclass_type, source):
             field_class, field_kwargs = self.build_property_field(source)
@@ -317,6 +344,9 @@ class DataclassSerializer(rest_framework.serializers.Serializer):
             field_kwargs['required'] = False
             field_kwargs['allow_null'] = True
 
+        if type_info.is_final:
+            field_kwargs['read_only'] = True
+
         return field_class, field_kwargs
 
     def build_composite_field(self, field_name: str, type_info: TypeInfo,
@@ -324,6 +354,15 @@ class DataclassSerializer(rest_framework.serializers.Serializer):
         """
         Create a composite (mapping or list) field.
         """
+        if type_info.is_mapping:
+            field_class = rest_framework.fields.DictField
+        else:
+            field_class = rest_framework.fields.ListField
+
+        # If the base type is not specified or is any type, we don't have to bother creating the child field.
+        if type_info.base_type is None:
+            return field_class, {}
+
         # Recurse to build the child field (i.e. the field of every instance). We pass the extra kwargs that are
         # specified for the child field through, so these can be used to recursively specify kwargs for child fields.
         extra_child_field_kwargs = extra_kwargs.get('child_kwargs', {})
@@ -337,12 +376,6 @@ class DataclassSerializer(rest_framework.serializers.Serializer):
         # Create child field and initialize parent field kwargs
         child_field = child_field_class(**child_field_kwargs)
         field_kwargs = {'child': child_field}
-
-        if type_info.is_mapping:
-            field_class = rest_framework.fields.DictField
-        else:
-            field_class = rest_framework.fields.ListField
-
         return field_class, field_kwargs
 
     def build_standard_field(self, field_name: str, type_info: TypeInfo) -> SerializerFieldDefinition:
@@ -360,8 +393,9 @@ class DataclassSerializer(rest_framework.serializers.Serializer):
             field_type = self.dataclass_definition.field_types[field_name]
             raise NotImplementedError(
                 "Automatic serializer field deduction not supported for field '{field}' on '{dataclass}' "
-                "of type '{type}'."
-                .format(dataclass=self.dataclass_definition.dataclass_type.__name__, field=field_name, type=field_type)
+                "of type '{type}' (during search for field of type '{reduced_type}')."
+                .format(dataclass=self.dataclass_definition.dataclass_type.__name__, field=field_name,
+                        type=field_type, reduced_type=type_info.base_type)
             )
 
     def build_relational_field(self, field_name: str, type_info: TypeInfo) -> SerializerFieldDefinition:
@@ -482,6 +516,20 @@ class DataclassSerializer(rest_framework.serializers.Serializer):
 
         return extra_kwargs
 
+    # Methods to convert between internal normalized value and serialized representation.
+
+    def to_internal_value(self, data: Dict[str, Any]) -> T:
+        """
+        Convert a dictionary representation of the dataclass containing only primitive values to a dataclass instance.
+        """
+        native_values = super(DataclassSerializer, self).to_internal_value(data)
+        empty_values = {key: empty for key in self.dataclass_definition.fields.keys() if key not in native_values}
+
+        dataclass_type = self.dataclass_definition.dataclass_type
+        instance = dataclass_type(**native_values, **empty_values)
+
+        return instance
+
 
 class HyperlinkedDataclassSerializer(DataclassSerializer):
     serializer_related_field = HyperlinkedRelatedField
@@ -492,7 +540,6 @@ class HyperlinkedDataclassSerializer(DataclassSerializer):
         """
         field_class = HyperlinkedDataclassSerializer
         field_kwargs = {'dataclass': type_info.base_type,
-                        'many': type_info.is_many,
-                        'allow_null': type_info.is_optional}
+                        'many': type_info.is_many}
 
         return field_class, field_kwargs
